import subprocess
import sys

import tenacity
import yaml


class ApiError(Exception):
    pass


@tenacity.retry(reraise=True, retry=tenacity.retry_if_exception_type(ApiError), stop=tenacity.stop_after_attempt(3))
def get(kind, name=None):
    command = ['kubectl', 'get', '-o', 'yaml', kind]
    if name:
        command.append(name)
    get_process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=sys.stderr)
    objects = yaml.safe_load(get_process.stdout)
    if get_process.wait() != 0:
        raise ApiError
    return objects


def get_failed_pod_for_job(job_name):
    command = [
        'kubectl', 'get', 'pod',
        '-o', 'yaml',
        '--field-selector', 'status.phase=Failed',
        '-l', 'job-name={}'.format(job_name),
    ]
    get_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=sys.stderr)
    objects = yaml.safe_load(get_process.stdout)
    if get_process.wait() != 0:
        raise ApiError
    return objects


def logs(pod_name, container_name=None):
    command = ['kubectl', 'logs', pod_name]
    if container_name:
        command += ['-c', container_name]
    log_process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = log_process.communicate()
    if log_process.returncode != 0:
        raise ApiError
    return stdout, stderr


def create(definition):
    run_command_with_yaml_on_stdin(
        ['kubectl', 'create', '-f', '-'], definition)


def apply(definition):
    run_command_with_yaml_on_stdin(
        ['kubectl', 'apply', '--record', '-f', '-'], definition)


def replace(definition):
    run_command_with_yaml_on_stdin(
        ['kubectl', 'replace', '--force', '--cascade', '-f', '-'], definition)


def rolling_update(definition, name):
    run_command_with_yaml_on_stdin(
        ['kubectl', 'rolling-update', name, '-f', '-'],
        definition)


def delete(kind, name):
    subprocess.check_call(['kubectl', 'delete', kind, name])


def run_command_with_yaml_on_stdin(command, definition):
    create_process = subprocess.Popen(command, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = create_process.communicate(yaml.dump(definition).encode())
    create_process.stdin.close()
    if create_process.wait() != 0:
        raise ApiError(stderr)



#!/usr/bin/env python

import optparse
import pathlib

import yaml

from kubepy import appliers
from kubepy import appliers_options
from kubepy import base_commands


class InstallError(Exception):
    pass


class ApplyOneCommand(base_commands.BaseCommand):
    def get_optparser(self):
        parser = optparse.OptionParser(
            usage="usage: %prog [options] [job_name ...]",
            epilog="Installs or Executes selected definition",
        )
        parser.add_option(
            '--directory',
            dest='directories',
            action='append',
            help='installs definitions from this directory, can be defined multiple times to override definitions.',
        )
        parser.add_option(
            '--show-definition',
            dest='show_definition',
            action='store_true',
            default=False,
            help='shows definition instead of applying them.',
        )
        appliers_options.Options.add_applier_options(parser)
        return parser

    def handle(self, args, options):
        directory_strings = options.directories or ['.']
        directories = [pathlib.Path(directory_string).resolve() for directory_string in directory_strings]
        runner = appliers.DirectoriesApplier(directories, appliers_options.Options.from_parsed_options(options))
        if args:
            for job_name in args:
                if options.show_definition:
                    print(yaml.dump(runner.get_named_definition(job_name)))
                else:
                    runner.apply_named(job_name)
        else:
            raise base_commands.CommandError('Provide definition names.')


def run():
    ApplyOneCommand().run()


if __name__ == '__main__':
    run()



def parse_dict_options_callback(option, opt_str, value, parser):
    name, path = value.split('=', 1)
    getattr(parser.values, option.dest)[name] = path


class Options:
    PARSER_CONFIGURATION = [
        ('--build-tag', {
            'dest': 'build_tag', 'action': 'store', 'help': 'used image tag and name suffix',
        }),
        ('--label', {
            'dest': 'labels',
            'action': 'callback',
            'type': 'string',
            'callback': parse_dict_options_callback,
            'help': 'add label to definition. Format is key=value. Can be used multiple times.',
        }),
        ('--label-pod', {
            'dest': 'pod_labels',
            'action': 'callback',
            'type': 'string',
            'callback': parse_dict_options_callback,
            'help': 'add label to each pod definition. Format is key=value. Can be used multiple times.',
        }),
        ('--annotate', {
            'dest': 'annotations',
            'action': 'callback',
            'type': 'string',
            'callback': parse_dict_options_callback,
            'help': 'add annotation to definition. Format is key=value. Can be used multiple times.',
        }),
        ('--annotate-pod', {
            'dest': 'pod_annotations',
            'action': 'callback',
            'type': 'string',
            'callback': parse_dict_options_callback,
            'help': 'add annotation to each pod definition. Format is key=value. Can be used multiple times.',
        }),
        ('--replace', {
            'dest': 'replace',
            'action': 'store_true',
            'help': 'delete and recreate deployments instead of doing rolling update',
        }),
        ('--host-volume', {
            'dest': 'host_volumes',
            'action': 'callback',
            'type': 'string',
            'callback': parse_dict_options_callback,
            'help': 'add host volume to pods. Format is name=path  (--host-volume=dev-volume=/home)',
        }),
        ('--env', {
            'dest': 'environment',
            'action': 'callback',
            'type': 'string',
            'callback': parse_dict_options_callback,
            'help': 'add environment variable to containers. Format is VAR=value  (--env=BUILD_NUMBER=2)',
        }),
        ('--max-job-retries', {
            'dest': 'max_job_retries',
            'action': 'store',
            'type': 'int',
            'help': 'When applying job fail if job fails n times.',
        }),
    ]

    def __init__(self, *, build_tag='latest', labels=None, pod_labels=None, annotations=None, pod_annotations=None,
                 replace=False, host_volumes=None, environment=None, max_job_retries=None):
        self.build_tag = build_tag
        self.labels = labels or {}
        self.pod_labels = pod_labels or {}
        self.annotations = annotations or {}
        self.pod_annotations = pod_annotations or {}
        self.replace = replace
        self.host_volumes = host_volumes or {}
        self.environment = environment or {}
        self.max_job_retries = max_job_retries

    @classmethod
    def add_applier_options(cls, parser):
        default_options = cls()
        for option, option_kwargs in cls.PARSER_CONFIGURATION:
            attribute = option_kwargs['dest']
            parser.add_option(option, default=getattr(default_options, attribute), **option_kwargs)

    @classmethod
    def from_parsed_options(cls, parsed_options):
        kwargs = {}
        for _, option_kwargs in cls.PARSER_CONFIGURATION:
            attribute = option_kwargs['dest']
            kwargs[attribute] = getattr(parsed_options, attribute)
        return cls(**kwargs)



import copy


def tag_untaged_images(definition, tag):
    new_definition = copy.deepcopy(definition)
    for container in iterate_container_definitions(new_definition):
        container['image'] = tag_untaged_image(container['image'], tag)
    return new_definition


def set_environment(definition, new_environment):
    new_definition = copy.deepcopy(definition)
    for container in iterate_container_definitions(new_definition):
        env_definition = container.get('env', [])
        env_definition = [
            env for env in env_definition if env['name'] not in new_environment.keys()
        ]
        for name, value in new_environment.items():
            env_definition.append({'name': name, 'value': value})
        container['env'] = env_definition
    return new_definition


def iterate_container_definitions(new_definition):
    return get_crawler(new_definition).get_container_definitions()


def add_host_volumes(definition, host_volumes):
    new_definition = copy.deepcopy(definition)
    for name, path in host_volumes.items():
        volumes = get_crawler(new_definition).get_volume_definitions()
        volumes.append({'name': name, 'hostPath': {'path': str(path)}})
    return new_definition


def add_labels(definition, labels, pod_labels):
    new_definition = copy.deepcopy(definition)
    crawler = get_crawler(new_definition)
    add_labels_to_metadata(metadata=crawler.get_metadata_definition(), labels=labels)
    add_labels_to_metadata(metadata=crawler.get_pod_metadata_definition(), labels=pod_labels)
    return new_definition


def add_labels_to_metadata(metadata, labels):
    new_labels = metadata.get('labels', {})
    for name, value in labels.items():
        new_labels[name] = value
    metadata['labels'] = new_labels


def add_annotations(definition, annotations, pod_annotations):
    new_definition = copy.deepcopy(definition)
    crawler = get_crawler(new_definition)
    add_annotations_to_metadata(metadata=crawler.get_metadata_definition(), annotations=annotations)
    add_annotations_to_metadata(metadata=crawler.get_pod_metadata_definition(), annotations=pod_annotations)
    return new_definition


def add_annotations_to_metadata(metadata, annotations):
    new_annotations = metadata.get('annotations', {})
    for name, value in annotations.items():
        new_annotations[name] = value
    metadata['annotations'] = new_annotations


def get_crawler(definition):
    return CRAWLER_CLASS_MAP[definition['kind']](definition)


class BaseCrawler:
    def __init__(self, definition):
        self.definition = definition

    def get_container_definitions(self):
        return self.get_pod_spec()['containers']

    def get_volume_definitions(self):
        return self.get_pod_spec().setdefault('volumes', [])

    def get_metadata_definition(self):
        return self.definition.setdefault('metadata', {})

    def get_pod_metadata_definition(self):
        return self.get_pod_metadata()

    def get_pod_spec(self):
        raise NotImplementedError

    def get_pod_metadata(self):
        raise NotImplementedError


class CronJobCrawler(BaseCrawler):
    def get_pod_spec(self):
        return self.get_job_crawler().get_pod_spec()

    def get_pod_metadata(self):
        return self.get_job_crawler().get_pod_metadata()

    def get_job_crawler(self):
        return DefinitionWithPodTemplateCrawler(self.definition['spec']['jobTemplate'])


class DefinitionWithPodTemplateCrawler(BaseCrawler):
    def get_pod_spec(self):
        return self.get_pod_crawler().get_pod_spec()

    def get_pod_metadata(self):
        return self.get_pod_crawler().get_pod_metadata()

    def get_pod_crawler(self):
        return PodCrawler(self.definition['spec']['template'])


class PodCrawler(BaseCrawler):
    def get_pod_spec(self):
        return self.definition['spec']

    def get_pod_metadata(self):
        return self.definition.setdefault('metadata', {})


def tag_untaged_image(image, tag):
    if ':' in image:
        return image
    else:
        return image + ':' + tag


CRAWLER_CLASS_MAP = {
    'Job': DefinitionWithPodTemplateCrawler,
    'CronJob': CronJobCrawler,
    'Deployment': DefinitionWithPodTemplateCrawler,
    'StatefulSet': DefinitionWithPodTemplateCrawler,
    'Pod': PodCrawler,
}



import collections
import copy
import logging
import time

from kubepy import api
from kubepy import definition_manager
from kubepy import definition_transformers

logger = logging.getLogger(__name__)


class InstallError(Exception):
    pass


class JobError(InstallError):
    pass


class PodError(JobError):
    def __init__(self, message='', container_name='', stdout='', stderr=''):
        self.container_name = container_name
        self.stdout = stdout
        self.stderr = stderr
        super().__init__(message, container_name, stdout, stderr)


class DeadlineExceeded(JobError):
    pass


def directory_applier(path, options):
    manager = definition_manager.DefinitionManager(path)
    return DefinitionsApplier(manager, options)


DirectoryApplier = directory_applier


def directories_applier(paths, options):
    manager = definition_manager.OverridenDefinitionManager(
        *(definition_manager.DefinitionManager(path) for path in paths),
    )
    return DefinitionsApplier(manager, options)


DirectoriesApplier = directories_applier


class DefinitionsApplier:
    def __init__(self, manager, options):
        self.options = options
        self.manager = manager

    def apply_all(self):
        for definition in self.manager.values():
            UniversalDefinitionApplier(definition, self.options).apply()

    def apply_named(self, name):
        definition = self.manager[name]
        UniversalDefinitionApplier(definition, self.options).apply()

    def get_named_definition(self, name):
        definition = self.manager[name]
        return UniversalDefinitionApplier(definition, self.options).new_definition


class BaseDefinitionApplier:
    def __init__(self, definition, options):
        self.definition = definition
        self.options = options

    def apply(self):
        raise NotImplementedError

    @property
    def usable_with(self):
        raise NotImplementedError

    @property
    def new_definition(self):
        raise NotImplementedError


class ResourceApplier(BaseDefinitionApplier):
    usable_with = ['Service', 'Secret', 'StorageClass', 'PersistentVolume',
                   'PersistentVolumeClaim', 'Ingress', 'PodDisruptionBudget']

    def apply(self):
        api.apply(self.definition)


class ReplicatedTemplateResourceApplier(BaseDefinitionApplier):
    usable_with = ['Deployment', 'StatefulSet']

    def apply(self):
        if self.options.replace:
            api.replace(self.new_definition)
        else:
            api.apply(self.new_definition)

    @property
    def new_definition(self):
        return transform_pod_definition(self.definition, self.options)


class CronJobApplier(BaseDefinitionApplier):
    usable_with = ['CronJob']

    def apply(self):
        api.apply(self.new_definition)

    @property
    def new_definition(self):
        return transform_pod_definition(self.definition, self.options)


class BaseJobApplier(BaseDefinitionApplier):
    def apply(self):
        api.create(self.new_definition)
        try:
            while True:
                status = self._get_status()
                status.raise_if_failed()
                if status.succeeded:
                    break
                else:
                    time.sleep(1)
        finally:
            api.delete(self.definition_type, self.name)

    def _get_status(self):
        return self.status_class(self.name, self._get_raw_status())

    def _get_raw_status(self):
        return api.get(self.definition_type, self.name)['status']

    @property
    def new_definition(self):
        return transform_pod_definition(self.definition, self.options)

    @property
    def name(self):
        return self.definition['metadata']['name']

    @property
    def status_class(self):
        raise NotImplementedError

    @property
    def definition_type(self):
        raise NotImplementedError


class BaseJobStatus:
    def __init__(self, definition_name, status):
        self.definition_name = definition_name
        self.status = status

    def raise_if_failed(self):
        raise NotImplementedError

    def succeeded(self):
        raise NotImplementedError


class JobStatus(BaseJobStatus):
    def __init__(self, definition_name, status, max_retires=None):
        super().__init__(definition_name, status)
        self.max_retries = max_retires

    def raise_if_failed(self):
        if 'conditions' in self.status:
            condition = self.status['conditions'][0]
            if condition['type'] == 'Failed':
                self._raise_for_failed_condition(condition)
        failures = self.status.get('failed', 0)
        if self.max_retries and failures > self.max_retries:
            failed_pod = self.get_job_pod()
            if failed_pod:
                stdout = self.get_failed_pod_logs(failed_pod)
                print(stdout)
            raise JobError('Job failed {} times.'.format(failures))

    def get_failed_pod_logs(self, failed_pod):
        for init_container_status in failed_pod['status'].get('initContainerStatuses', []):
            if init_container_status['state'].get('terminated', {}).get('reason') == 'ContainerCannotRun':
                return 'Init container failed! ' + str(init_container_status['state']['terminated']['message'])
        for container_status in failed_pod['status'].get('containerStatuses', []):
            if container_status['state'].get('terminated', {}).get('reason') == 'ContainerCannotRun':
                return 'Container failed! ' + str(container_status['state']['terminated']['message'])
        return api.logs(failed_pod['metadata']['name'])[0].decode()

    def _raise_for_failed_condition(self, condition):
        if condition['reason'] == 'DeadlineExceeded':
            raise DeadlineExceeded(condition['message'])
        else:
            raise JobError(condition['message'])

    @property
    def succeeded(self):
        return 'completionTime' in self.status

    def get_job_pod(self):
        pods = api.get_failed_pod_for_job(self.definition_name)['items']
        if not pods:
            logging.warning('No pod found for job {}!'.format(self.definition_name))
            return
        elif len(pods) > 1:
            logging.info('More than one pod found for job {}, returning first.'.format(self.definition_name))
        return pods[0]


ContainerInfo = collections.namedtuple('ContainerInfo', ['name', 'state'])


class PodStatus(BaseJobStatus):
    def raise_if_failed(self):
        for container in self.containers:
            if 'terminated' in container.state:
                if container.state['terminated']['reason'] != 'Completed':
                    self.raise_with_log(container.name)

    @property
    def succeeded(self):
        for container in self.containers:
            terminated = 'terminated' in container.state
            completed = terminated and container.state['terminated']['reason'] == 'Completed'
            if not completed:
                return False
        else:
            return True

    @property
    def containers(self):
        for container_status in self.status['containerStatuses']:
            yield ContainerInfo(container_status['name'], container_status['state'])

    def raise_with_log(self, container_name):
        stdout, stderr = api.logs(self.definition_name, container_name)
        raise PodError('Failure in {}'.format(container_name), container_name, stdout, stderr)


class JobApplier(BaseJobApplier):
    definition_type = 'Job'
    usable_with = [definition_type]
    status_class = JobStatus

    def _get_status(self):
        return self.status_class(self.name, self._get_raw_status(), self.options.max_job_retries)


class PodApplier(BaseJobApplier):
    definition_type = 'Pod'
    usable_with = [definition_type]
    status_class = PodStatus

    def apply(self):
        if self.definition['spec']['restartPolicy'] != 'Never':
            raise JobError('Pod has to have restartPolicy = Never')
        super().apply()


class UniversalDefinitionApplier(BaseDefinitionApplier):
    applier_classes = (ResourceApplier, ReplicatedTemplateResourceApplier, CronJobApplier, JobApplier, PodApplier)
    usable_with = sum((applier.usable_with for applier in applier_classes), [])

    def apply(self):
        self.get_applier().apply()

    @property
    def kind_map(self):
        kind_map = UniqueDict()
        for applier_class in self.applier_classes:
            for kind in applier_class.usable_with:
                kind_map[kind] = applier_class
        return dict(kind_map)

    @property
    def new_definition(self):
        return self.get_applier().new_definition

    def get_applier(self):
        kind = self.definition['kind']
        try:
            applier_class = self.kind_map[kind]
        except KeyError:
            raise InstallError('Unknown resource kind: {}'.format(kind))
        else:
            return applier_class(self.definition, self.options)


class UniqueDictException(Exception):
    pass


class UniqueDict(dict):
    def __setitem__(self, key, value):
        if key not in self:
            super().__setitem__(key, value)
        else:
            raise UniqueDictException(key)


def transform_pod_definition(definition, options):
    new_definition = copy.deepcopy(definition)
    new_definition = definition_transformers.tag_untaged_images(new_definition, options.build_tag)
    new_definition = definition_transformers.set_environment(new_definition, options.environment)
    new_definition = definition_transformers.add_host_volumes(new_definition, options.host_volumes)
    new_definition = definition_transformers.add_labels(new_definition, options.labels, options.pod_labels)
    new_definition = definition_transformers.add_annotations(
        new_definition, options.annotations, options.pod_annotations)
    return new_definition



try:
    from pip._internal.req import parse_requirements
except ImportError:
    from pip.req import parse_requirements
from setuptools import find_packages
from setuptools import setup


def get_long_description():
    with open('README.md') as readme_file:
        return readme_file.read()


setup(
    name='kubepy',
    version='1.14.0',
    description='Python wrapper on kubectl that makes deploying easy.',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Social WiFi',
    author_email='it@socialwifi.com',
    url='https://github.com/socialwifi/kubepy',
    packages=find_packages(exclude=['tests']),
    install_requires=[str(ir.req) for ir in parse_requirements('base_requirements.txt', session=False)],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'kubepy-apply-all = kubepy.commands.apply_all:run',
            'kubepy-apply-one = kubepy.commands.apply_one:run',
        ],
    },
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
